/*****************************************************************************
/*
                                  WASD.h

Main header file for WASD VMS HTTP daemon.


VERSION HISTORY
---------------
08-JUN-97  MGD  v4.2.0, CGIplus
01-FEB-97  MGD  v4.0.0, version 4 :^)
01-AUG-96  MGD  v3.3.0, quite a few changes since 3.0.n
01-DEC-95  MGD  v3.0.0, HTTPd version 3
20-DEC-94  MGD  v2.0.0, multi-threaded version
20-JUN-94  MGD  v1.0.0, single-threaded version
*/
/*****************************************************************************/

#include <descrip.h>
#include <rms.h>
#include <rmsdef.h>

#ifndef WASD_AUTH_H
#include "auth.h"
#endif

#ifdef __ALPHA
#   pragma nomember_alignment
#endif

/**********/
/* macros */
/**********/

#define boolean int
#define true 1
#define false 0

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) (!((x) & STS$M_SUCCESS))

#define AuthConfigFileName "HTTPD$AUTH"
#define ConfigFileName "HTTPD$CONFIG"
#define MapFileName "HTTPD$MAP"

/* this value must also include connections in a "keep-alive" state */
#define DefaultMaxConcurrentConnections 20

#define DefaultInputTimeoutMinutes 2
#define DefaultOutputTimeoutMinutes 10
#define DefaultKeepAliveTimeoutSeconds 5

#define DefaultDclSysOutputSize 4096
#define DefaultFileBufferSize 1024
#define DefaultNetReadBufferSize 2048
#define DefaultOutputBufferSize 4096

#define DefaultDirListingLayout "I__L__R__S__D"

#define HttpdInternalScriptUpd "/upd"
#define HttpdInternalScriptTree "/tree"

#define HttpdInternalPasswordChange "/httpd/-/change/"

#define HttpdInternal "/httpd/-/"

#define HttpdInternalAdmin "/httpd/-/admin/"

#define HttpdInternalControl "/httpd/-/admin/control/"
#define HttpdInternalControlRestart "/httpd/-/admin/control/restart"
#define HttpdInternalControlExit "/httpd/-/admin/control/exit"
#define HttpdInternalControlAbort "/httpd/-/admin/control/abort"
#define HttpdInternalControlLogOn "/httpd/-/admin/control/logon"
#define HttpdInternalControlLogOff "/httpd/-/admin/control/logoff"
#define HttpdInternalControlLogFlush "/httpd/-/admin/control/logflush"

#define HttpdInternalGraphic "/httpd/-/admin/graphic/"
#define HttpdInternalGraphicActivity "/httpd/-/admin/graphic/activity"

#define HttpdInternalReport "/httpd/-/admin/report/"
#define HttpdInternalReportActivity "/httpd/-/admin/report/activity"
#define HttpdInternalReportRules "/httpd/-/admin/report/rules"

#define HttpdInternalRevise "/httpd/-/admin/revise/"
#define HttpdInternalReviseAuth "/httpd/-/admin/revise/auth/"

/* helps ensure a client doesn't get into an infinite loop */
#define MaxKeepAliveCount 100

/*
   Attempt to reduce dynamic memory fragmentation within C RTL by
   maintaining a minimum size chunk to be allocated.  Hopefully this
   will prevent lots of small chunks having to be garbage-collected
   or otherwise processed (as has sometimes been observed ... at least
   this is the best explanation I can come up with).
*/
#define HEAP_MIN_CHUNK_SIZE 256

/* control pre-expired responses */
#define PRE_EXPIRE_ADMIN 1
#define PRE_EXPIRE_DELETE_ON_CLOSE 1
#define PRE_EXPIRE_INDEX_OF 0
#define PRE_EXPIRE_MENU 1
#define PRE_EXPIRE_PUT 1
#define PRE_EXPIRE_SSI 0
#define PRE_EXPIRE_UPD 1
#define PRE_EXPIRE_UPD_TREE_OF 1

/* constants used to identify HTTP methods */
#define HTTP_METHOD_DELETE 0x00000001
#define HTTP_METHOD_GET 0x00000002
#define HTTP_METHOD_HEAD 0x00000004
#define HTTP_METHOD_POST 0x00000008
#define HTTP_METHOD_PUT 0x00000010

/* constants for controlling logging */
#define LOGGING_BEGIN 1
#define LOGGING_END 2
#define LOGGING_OPEN 3
#define LOGGING_CLOSE 4
#define LOGGING_FLUSH 5
#define LOGGING_ENTRY 6

/* constants used by input/output timeout functionality */
#define TIMEOUT_INITIALIZE 0
#define TIMEOUT_INPUT 1
#define TIMEOUT_OUTPUT 2
#define TIMEOUT_KEEPALIVE 3

/* constants used by the description module */
#define DESCRIPTION_ALL 0x01
#define DESCRIPTION_TEXT_PLAIN 0x02
#define DESCRIPTION_TEXT_HTML 0x04
/* this is returned in the first character the file type contains none */
#define DESCRIPTION_IMPOSSIBLE 0xff

/* if non-zero then imposes a maximum byte count on a request body */
#define MaxContentLength 0

/* maximum Kbytes (1024) that can be in a body to be POSTed or PUTted */
#define DefaultPutMaxKbytes 250
/* number of versions retained for files POSTed or PUTed */
#define DefaultPutVersionLimit 3

/* header fields for HTTP/1.0 keep alive response */
#define KeepAliveHttpHeader "Connection: Keep-Alive\r\nKeep-Alive:\r\n"

/* if true, stops HTTPd server account with SYSPRV PUT(etc)ing - DANGEROUS */
#define PUT_DISALLOW_SYSPRV 0

#define MaxConfigHomePages 16
#define MaxConfigReadMeFiles 16

#define SCRIPT_NAME_SIZE 128
#define METHOD_NAME_SIZE 16

/*******************/
/* data structures */
/*******************/

# pragma member_alignment __save
# pragma nomember_alignment

struct VmsItem
{
   unsigned short  buf_len;
   unsigned short  item;
   unsigned char   *buf_addr;
   unsigned short  *ret_len;
};

struct AnExitHandler
{
   unsigned long  Flink;
   unsigned long  HandlerAddress;
   unsigned long  ArgCount;
   unsigned long  ExitStatusPtr;
};

struct AnIOsb
{
   unsigned short  Status;
   unsigned short  Count;
   unsigned long  Unused;
};

struct MbxSenseIOsb
{
   unsigned short  Status;
   unsigned short  MessageCount;
   unsigned long  MessageBytes;
};

#pragma member_alignment __restore

/*
   Store these structures naturally-aligned on AXP.
   Uses a bit more storage but makes access as efficient as possible.
*/

#ifdef __ALPHA
#   pragma member_alignment __save
#   pragma member_alignment
#endif

/* per-request memory heap */
struct HeapStruct
{
   struct HeapStruct  *PrevPtr;
   struct HeapStruct  *NextPtr;
   unsigned int  ChunkSize;
   /* the actual data follows immediately after the length */
};

/*
This structure is used to both store and to point at the various components
depending on the context.
*/

struct ContentTypeStruct
{
   struct ContentTypeStruct  *NextPtr;
   char  *SuffixPtr;
   char  *ContentTypePtr;
   char  *LastFoundContentTypePtr;
   char  *AutoScriptNamePtr;
   char  *ContentDescriptionPtr;
   /* string data is stored here */
};

/*
This structure is used to both store and to point at the various components
depending on the context.
*/

struct IconStruct
{
   struct IconStruct  *NextPtr;
   struct IconStruct  *LastFoundIconPtr;
   /* string data is stored here */
};

struct  GenericListEntryStruct
{
   struct GenericListEntryStruct  *PrevPtr;
   struct GenericListEntryStruct  *NextPtr;
   void  *DataPtr;
};

struct  GenericListStruct
{
   struct GenericListEntryStruct  *HeadPtr;
   struct GenericListEntryStruct  *TailPtr;
   int  EntryCount;
};

/**********************************/
/* request history data structure */
/**********************************/

/* 256 bytes (a somewhat arbitrary quantity) */
struct  RequestHistoryStruct
{
   struct  GenericListEntryStruct  HistoryListEntry;
   unsigned long  BinaryTime [2];
   unsigned long  BytesRx,
                  BytesTx;
   int  ResponseStatusCode;
   char  *RequestPtr;
   boolean  Truncated;
   char  ClientAndRequest [256-12-8-4-4-4-4-4];
};

/**********************************/
/* server configuration structure */
/**********************************/

struct ConfigStruct
{
   boolean  AuthBasicEnabled,
            AuthDigestEnabled,
            AuthVmsEnabled,
            DclFullRequest,
            DclSpawnAuthPriv,
            DirAccess,
            DirAccessSelective,
            DirPreExpired,
            DirOwnerEnabled,
            DirReadMeBottom,
            DirReadMeTop,
            DirWildcardEnabled,
            DnsLookup,
            ErrorRecommend,
            IncludeCommentedInfo,
            LoggingEnabled,
            MonitorEnabled,
            SsiEnabled,
            SsiAccessesEnabled,
            SsiExecEnabled;

   int  AcceptHostsLength,
        ActivityNumberOfDays,
        AuthDigestNonceGetLifeTime,
        AuthDigestNoncePutLifeTime,
        Busy,
        DclCgiPlusLifeTime,
        DclSubprocessHardLimit,
        DclSubprocessSoftLimit,
        DclZombieLifeTime,
        HomePageCount,
        InputTimeoutMinutes,
        KeepAliveTimeoutSeconds,
        OutputTimeoutMinutes,
        PutMaxKbytes,
        PutVersionLimit,
        ReadMeFileCount,
        RejectHostsLength,
        RequestHistory,
        ServerPort,
        StreamLfConversionMaxKbytes;

   char  DefaultDirLayout [32],
         KeywordSearch [128];

   char  *AcceptHostsPtr,
         *HomePageArray [MaxConfigHomePages],
         *DefaultPost,
         *ErrorInfoPtr,
         *ErrorSysAdminPtr,
         *HtmlContentType,
         *IsMapContentType,
         *MenuContentType,
         *PlainTextContentType,
         *ReadMeFileArray [MaxConfigReadMeFiles],
         *RejectHostsPtr,
         *SsiContentType;

   struct ContentTypeStruct  *ContentTypeListHeadPtr,
                             *ContentTypeListTailPtr;
   int  ContentTypeStructOverhead;

   struct IconStruct  *IconListHeadPtr,
                      *IconListTailPtr;
   int  IconStructOverhead;

   /* only used while reading configuration from file */
   int  LineNumber,
        ProblemCount,
        ProblemReportLength;
   char  *ProblemReportPtr;
   char  DirectiveName [32],
         LoadFileName [256];
   unsigned long  LoadBinTime [2],
                  RevBinTime [2];

   /* only used when reporting or changing configuration */
   void  *RequestPtr;
};

/*************************/
/* per-server accounting */
/*************************/

struct AccountingStruct
{
   /* this MUST be the first member to allow zeroing of the rest */
   unsigned long  ZeroedCount;

   unsigned long  AuthBasicCount,
                  AuthAuthorizedCount,
                  AuthDigestCount,
                  AuthHtDatabaseCount,
                  AuthNotAuthenticatedCount,
                  AuthNotAuthorizedCount,
                  AuthVmsCount,
                  ConnectAcceptedCount,
                  ConnectCallocFailCount,
                  ConnectCount,
                  ConnectCurrent,
                  ConnectPeak,
                  ConnectRejectedCount,
                  ConnectTooBusyCount,
                  DclCgiPlusReusedCount,
                  DclSpawnCount,
                  DoAutoScriptCount,
                  DoCgiPlusScriptCount,
                  DoDclCommandCount,
                  DoDirectoryCount,
                  DoIsMapCount,
                  DoScriptCount,
                  DoFileCount,
                  DoFileNotModifiedCount,
                  DoMenuCount,
                  DoPutCount,
                  DoServerAdminCount,
                  DoSsiCount,
                  DoUpdateCount,
                  HeapCallocFailCount,
                  HeapReallocFailCount,
                  InternalErrorCount,
                  LastExitStatus,
                  MethodDeleteCount,
                  MethodGetCount,
                  MethodHeadCount,
                  MethodPostCount,
                  MethodPutCount,
                  MethodTraceCount,
                  RedirectLocalCount,
                  RedirectRemoteCount,
                  RequestTotalCount,
                  RequestErrorCount,
                  RequestForbiddenCount,
                  RequestKeepAliveCount,
                  RequestKeepAliveMax,
                  RequestParseCount,
                  RequestPragmaNoCacheCount,
                  RequestProxyCount,
                  RequestRmsErrorCount,
                  StartupCount,
                  StreamLfConversionCount;

   unsigned long  QuadBytesRx [2],
                  QuadBytesTx [2];

   unsigned long  ResponseStatusCodeCountArray [6];
};

/***********************/
/* graph/gif structure */
/***********************/

struct GraphTaskStruct
{
   /* the structure in actually defined in Graph.c */
   void  *GraphPtr;

   /* pointer a request using this graph structure */
   struct RequestStruct  *rqptr;

   /* pointer to function, used for specifying the next task */
   void (*NextTaskFunction)(struct RequestStruct*);
};

/************************/
/* DCL module structure */
/************************/

struct DclTaskStruct
{
   boolean  /* add a newline to each record from a script subprocess */
            AddNewline,
            /* record-by-record output from DCL, flush immediately */
            BufferRecords,
            /* when script subprocess output still within HTTP header */
            CheckSubprocessResponseHeader,
            /* this is a "standard" CGI script */
            IsCgiScript,
            /* this is a "CGI-plus" script */
            IsCgiPlusScript,
            /* this is merely a DCL command (from SSI) */
            IsDclCommand,
            /* when the task completes the subprocess is deleted */
            IsMarkedForDelete;
            
   int  /* number of times the CGIplus script has been used */
        CgiPlusUsageCount,
        /* remaining content in POST script */
        ContentLength,
        /* length of the DCL EOF string */
        DclEofLength,
        /* count of I/Os CGIplus variables reads still outstanding */
        QueuedCgiPlusIn,
        /* count of HTTP$INPUT reads still outstanding */
        QueuedHttpInput,
        /* count of I/Os between subprocess and HTTPd still outstanding */
        QueuedSysCommand,
        /* when the count of I/Os is considered task finished */
        QueuedSysCommandAllowed,
        /* count of I/Os between subprocess and HTTPd still outstanding */
        QueuedSysOutput,
        /* size of DCL subprocess' SYS$OUTPUT buffer */
        SysOutputSize,
        /* total number of times the task slot was used */
        TotalUsageCount,
        /* number of subprocess has been a zombie */
        ZombieCount;

   unsigned short  /* channel to subprocess' CGIPLUSIN mailbox */
                   CgiPlusInChannel,
                   /* channel to subprocess' HTTP$INPUT mailbox */
                   HttpInputChannel,
                   /* channel to subprocess' SYS$INPUT mailbox */
                   SysCommandChannel,
                   /* channel to subprocess' SYS$OUTPUT mailbox */
                   SysOutputChannel;

   unsigned long  /* subprocess' exit status */
                  SubprocessCompletionStatus,
                  /* subprocess' PID */
                  SubprocessPid;

   unsigned long  /* last time CGIplus script was used */
                  LastUsedBinaryTime [2];

   unsigned char  /* content pointer in POST script */
                  *ContentPtr,
                  /* heap storage for buffering subprocess' SYS$OUTPUT */
                  *SysOutputPtr;
                  
   unsigned char  /* device name of subprocess' CGIPLUSIN mailbox */
                  CgiPlusInDevName [64],
                  /* DCL command (from SSI) */
                  DclCommand [256],
                  /* DCL EOF string */
                  DclEof [64],
                  /* device name of subprocess' HTTP input mailbox */
                  HttpInputDevName [64],
                  /* script name */
                  ScriptName [SCRIPT_NAME_SIZE],
                  /* device name of subprocess' input mailbox */
                  SysCommandDevName [64],
                  /* device name of subprocess' output mailbox */
                  SysOutputDevName [64];

   struct AnIOsb  CgiPlusInIOsb,
                  HttpInputIOsb,
                  SysCommandIOsb,
                  SysOutputIOsb;

   /* string descriptors for the various mailbox device names */
   struct dsc$descriptor_s  CgiPlusInDevNameDsc;
   struct dsc$descriptor_s  HttpInputDevNameDsc;
   struct dsc$descriptor_s  SysCommandDevNameDsc;
   struct dsc$descriptor_s  SysOutputDevNameDsc;

   /* pointer a request using this DCL task structure */
   struct RequestStruct  *rqptr;

   /* pointer to function, used for specifying the next task */
   void (*NextTaskFunction)(struct RequestStruct*);

   /* */
   struct  GenericListEntryStruct  DclListEntry;
};

/********************************/
/* description module structure */
/********************************/

struct DescrTaskStruct
{
   boolean  /* getting description from HTML, inside <TITLE></TITLE> tag */
            DescriptionInside,
            /* a <TITLE></TITLE> description has been retrieved from HTML */
            DescriptionRetrieved;

   int  /* internal description, number of lines checked from each file */
        DescriptionLineCount,
        /* internal description, number of "<" encountered */
        DescriptionClosingTagCount,
        /* internal description, number of ">" encountered */
        DescriptionOpeningTagCount,
        /* */
        FileNameLength,
        /* */
        SizeOfDescriptionBuffer,
        /* buffer for most recent task status */
        TaskStatusBuffer;

   unsigned char  /* pointer */
                  *DescriptionPtr,
                  /* pointer */
                  *DescriptionBufferPtr;

   char  /* space for generating file's description */
         Description [256],
         /* */
         FileName [256],
         /* */
         FilePath [256],
         /* */
         ReadBuffer [256];

   struct FAB  FileFab;
   struct RAB  FileRab;

   struct DescrTaskStruct  *DescrTaskPtr;

   /* pointer to function, used for specifying the next task */
   void (*NextTaskFunction)(struct RequestStruct*);
};

/******************************/
/* directory module structure */
/******************************/

struct DirTaskStruct
{
   boolean  /* disables auto-scripting appending version */
            AutoScriptEnabled,
            /* present a directory listing ("Index of") in a VMS-style */
            FormatLikeVms,
            /* the original specification's file/type/version was not empty */
            DirSpecIncludedFilePart,
            /* */
            ParseInUse;

   int  /* count of directories in listing */
        DirectoryCount,
        /* */
        LayoutFaoLength, 
        /* */
        LayoutHeadingLength, 
        /* */
        DirSpecLength,
        /* */
        FieldWidthCdt,
        /* */
        FieldWidthName,
        /* */
        FieldWidthOwner,
        /* */
        FieldWidthRdt,
        /* */
        FieldWidthSize,
        /* count of number of files in listing */
        FileCount,
        /* */
        ReadMeFileIndex,
        /* buffer for most recent task status */
        TaskStatusBuffer;

   unsigned long  /* for directory module, file's allocated blocks */
                  TotalAllocatedBlocks,
                  /* for directory module, file's used blocks */
                  TotalUsedBlocks;

   unsigned char  /* pointer to heap storage URL equivalent of directory */
                  *DirectoryPathPtr,
                  /* pointer to column headings string */
                  *LayoutHeadingPtr,
                  /* pointer to directory layout string */
                  *LayoutPtr,
                  /* pointer to sys$fao() directive string */
                  *LayoutFaoPtr;
                  
   char  /* */
         Description [256],
         /* directory part of file specification */
         DirectoryPart [128],
         /* directory specification */
         DirPath [256],
         /* directory path specification */
         DirSpec [256],
         /* */
         ExpandedFileName [256],
         /* file part of file specification */
         FilePart [128],
         /* query string for directory module */
         QueryString [128],
         /* */
         ResultantFileName [256],
         /* prepended script name */
         ScriptName [SCRIPT_NAME_SIZE];

   /* */
   struct ContentTypeStruct  ContentType;

   /* */
   struct IconStruct  IconUsed;

   /* RMS structures */
   struct FAB  FileFab;
   struct RAB  FileRab;
   struct NAM  FileNam;
   struct FAB  SearchFab;
   struct NAM  SearchNam;

   struct DirTaskStruct  *DirTaskPtr;

   /* pointer to function, used for specifying the next task */
   void (*NextTaskFunction)(struct RequestStruct*);
};

/*************************/
/* file module structure */
/*************************/

struct FileTaskStruct
{
   struct FileTaskStruct  *FileTaskPtr;

   boolean  /* */
            ParseInUse,
            /* to be read record-by-record, or by block I/O (from ACP QIO) */
            RecordIO,
            /* when true a file's contents are enclosed in <PRE></PRE> tags */
            PreTagFileContents,
            /* when a file's contents have been <PRE>ed, end with </PRE> tag */
            PreTagEndFileContents,
            /* when true HTML-forbidden characters are escaped */
            EscapeHtml;

   int  /* size of buffer for escaping HTML */
        EscapeHtmlSize,
        /* */
        FileNameLength,
        /* buffer for most recent task status */
        TaskStatusBuffer;

   unsigned long  /* (from ACP QIO) */
                  SizeInBytes;
   unsigned long  /* 64 bit VMS time when file was modified (from ACP QIO) */
                  ModifiedBinaryTime [2];

   char  /* */
         FileName [256];

   unsigned char  /* pointer to MIME content-type */
                  *ContentTypePtr,
                  /* pointer to buffer used to escape HTML-forbidden chars */
                  *EscapeHtmlPtr;

   /* RMS structures */
   struct FAB  FileFab;
   struct RAB  FileRab;
   struct NAM  FileNam;

   /* pointer to function, used for specifying the next task */
   void (*NextTaskFunction)(struct RequestStruct*);
};

/****************************/
/* HTAdmin module structure */
/****************************/

struct HTAdminTaskStruct
{
   struct HTAdminTaskStruct  *HTAdminTaskPtr;

   boolean  /* */
            BriefList,
            /* */
            ParseInUse;

   int  /* */
        FileCount,
        /* */
        FileNameLength,
        /* */
        RecordCount,
        /* */
        UserCount,
        /* */
        UserListCount,
        /* */
        UserListLength;

   char  /* */
         AuthFileName [64],
         /* */
         AuthFileSpec [64],
         /* */
         ExpandedFileName [256],
         /* */
         ResultantFileName [256];

   char  /* */
         *UserListPtr;

   struct AuthHtRecordStruct  AuthHtRecord;

   /* RMS structures */
   struct FAB  FileFab;
   struct RAB  FileRab;
   struct NAM  FileNam;
   struct XABDAT  FileXabDat;
   struct XABFHC  FileXabFhc;
   struct FAB  SearchFab;
   struct NAM  SearchNam;

   /* pointer to function, used for specifying the next task */
   void (*NextTaskFunction)(struct RequestStruct*);
};

/**************************/
/* IsMap module structure */
/**************************/

struct IsMapTaskStruct
{
   int  /* */
        FileNameLength,
        /* character number currently being parsed (for error information) */
        IsMapCharNumber,
        /* line number in current line (for error information) */
        IsMapLineNumber,
        /* buffer for most recent task status */
        TaskStatusBuffer;

   double  /* keeps track of closest 'point x,y' (if any) */
           IsMapClosestPoint,
           /* stores the numeric equivalent of the browser x,y click point */
           IsMapClickCoord [2];

   char  /* */
         FileName [256],
         /* */
         ReadBuffer [256];

   /* RMS structures */
   struct FAB  FileFab;
   struct RAB  FileRab;

   /* pointer to function, used for specifying the next task */
   void (*NextTaskFunction)(struct RequestStruct*);
};

/*************************/
/* menu module structure */
/*************************/

struct MenuTaskStruct
{
   boolean  /* */
            ParseInUse;

   int  /* */
        FileNameLength,
        /* used to count consecutive blank lines */
        MenuBlankLineCount,
        /* used to count consecutive non-blank lines */
        MenuLineCount,
        /* used to track current section number */
        MenuSectionNumber;

   unsigned char  /* pointer to heap storage of dynamic record buffer */
                  *MenuBufferPtr;

   char  /* */
         Description [256],
         /* sys$parse(), sys$search() storage */
         ExpandedFileName [256],
         /* */
         FileName [256],
         /* */
         ReadBuffer [256],
         /* sys$parse(), sys$search() storage */
         ResultantFileName [256];

   /* RMS structures */
   struct FAB  FileFab;
   struct RAB  FileRab;
   struct NAM  FileNam;
   struct XABDAT  FileXabDat;
   struct FAB  SearchFab;
   struct NAM  SearchNam;

   /* pointer to function, used for specifying the next task */
   void (*MenuAstFunctionPtr)(struct RequestStruct*);

   struct MenuTaskStruct  *MenuTaskPtr;

   /* pointer to function, used for specifying the next task */
   void (*NextTaskFunction)(struct RequestStruct*);
};

/************************/
/* put module structure */
/************************/

struct PutTaskStruct
{
   boolean  /* a file has been created by this request */
            FileCreated,
            /* document posted is "text/..." */
            ContentTypeText,
            /* document posted is URL-encoded */
            ContentTypeUrlEncoded,
            /* used when un-encoding URL-encoded posts */
            UrlEncodedInsideFormName;

   int  /* length of file contents pointed to in buffer */
        ContentFileLength,
        /* */
        FileCount,
        /* */
        FileSizeBytes,
        /* length of MIME boundary used to delimit mixed subparts in buffer */
        MultipartBoundaryLength,
        /* */
        FileContentLength,
        /* */
        FileNameLength;

   char  /* */
         ExpandedFileName [256],
         /* */
         FileName [256],
         /* */
         TemporaryFileName [256],
         /* */
         TemporaryFilePath [256];

   unsigned char  /* points to start of file contents in buffer */
                  *ContentFilePtr,
                  /* used as pointer into buffer */
                  *FileContentPtr,
                  /* used as pointer into buffer */
                  *FileContentTypePtr,
                  /* */
                  *FileNamePtr,
                  /* MIME boundary used to delimit mixed subparts in buffer */
                  *MultipartBoundaryPtr,
                  /* pointer to dynamic storage of directory files created in */
                  *MultipartDirectoryPtr,
                  /* */
                  *MultipartFileNamePtr,
                  /* */
                  *MultipartActionPtr;

   /* RMS structures */
   struct FAB  FileFab;
   struct RAB  FileRab;
   struct NAM  FileNam;
   struct XABDAT  FileXabDat;
   struct XABPRO  FileXabPro;

   struct PutTaskStruct  *PutTaskPtr;

   /* pointer to function used for specifying AST function following */
   void (*NextFunctionPtr)(struct RequestStruct*);

   /* pointer to function, used for specifying the next task */
   void (*NextTaskFunction)(struct RequestStruct*);

   /* pointer to function used for specifying AST function following */
   void (*WriteFileNextFunction)(struct RequestStruct*);
};

/************************/
/* SSI module structure */
/************************/

struct SsiTaskStruct
{
   int  /* line number where current directive began */
        DirectiveLineNumber,
        /* */
        FileNameLength,
        /* line number currently being parsed (for error information) */
        LineNumber,
        /* buffer for logical value */
        NetBufferOutputEscapeHtml,
        /* buffer for most recent task status */
        TaskStatusBuffer;
        
   char  /* */
         Directive [256],
         /* */
         FileName [256],
         /* */
         ReadBuffer [256];

   unsigned char  /* current character position in directive buffer */
                  *DirectivePtr,
                  /* current character position when parsing lines */
                  *LineParsePtr,
                  /* pointer to storage allocated for size format string */
                  *SizeFmtPtr,
                  /* pointer to storage allocated for time format string */
                  *TimeFmtPtr;
                                    
   struct FAB  FileFab;
   struct NAM  FileNam;
   struct RAB  FileRab;
   struct XABDAT  FileXabDat;
   struct XABPRO  FileXabPro;

   struct SsiTaskStruct  *SsiTaskPtr;

   /* pointer to function, used for specifying the next task */
   void (*NextTaskFunction)(struct RequestStruct*);
};

/***************************/
/* update module structure */
/***************************/

struct UpdTaskStruct
{
   struct UpdTaskStruct  *UpdTaskPtr;

   boolean  /* */
            BrowserFormHasOwnLineBreak,
            /* */
            BrowserSupportsFileUpload,
            /* */
            ConfigurationEdit,
            /* */
            FileExists,
            /* */
            ParseInUse,
            /* */
            RuleCheckAvailable,
            /* */
            UpdateTree;

   int  /* */
        Cols,
        /* */
        FileCount,
        /* */
        FileNameLength,
        /* */
        FilePathSize,
        /* */
        NavigateSize,
        /* */
        ParentDirLength,
        /* */
        Rows,
        /* */
        TreeLevel;

   char  /* */
         AsPath [256],
         /* */
         CxR [16],
         /* */
         ExpandedFileName [256],
         /* */
         FileName [256],
         /* tree processing, contains any name, type, version */
         FileNamePart [128],
         /* */
         FilePath [256],
         /* */
         HtmlEscapedTitle [256],
         /* */
         ParentDir [256],
         /* MUST be 512 bytes (one block) for file copy purposes */
         ReadBuffer [512],
         /* */
         ResultantFileName [256];

   unsigned char  /* pointer to MIME content-type */
                  *ContentTypePtr;

   /* RMS structures */
   struct FAB  FileFab;
   struct RAB  FileRab;
   struct NAM  FileNam;
   struct XABDAT  FileXabDat;
   struct XABFHC  FileXabFhc;
   struct FAB  SearchFab;
   struct NAM  SearchNam;

   struct TaskTreeStruct  *TreeHeadPtr,
                          *TreeNodePtr;

   /* pointer to function, used for specifying the next task */
   void (*NextTaskFunction)(struct RequestStruct*);
};

/********************************/
/* update module tree structure */
/********************************/

struct UpdTreeStruct
{
   struct UpdTreeStruct  *PrevTreeNodePtr,
                         *NextTreeNodePtr;

   int  /* */
        FileNameLength,
        /* */
        ParentDirLength;

   char  /* */
         ExpandedFileName [256],
         /* */
         FileName [256],
         /* */
         ParentDir [256],
         /* */
         ResultantFileName [256];

   /* RMS structures */
   struct FAB  SearchFab;
   struct NAM  SearchNam;
};

/*****************************/
/* per-thread data structure */
/*****************************/

struct RequestStruct
{
   /***************************/
   /* general purpose storage */
   /***************************/

   /*
      Storage areas at the beginning of this structure, specifically
      above the field 'RetainAboveZeroBelow', are retained between
      successive requests during persistent connections.
      When a "keep-alive" connection is reinitialized ready for a
      potential new request all of this structure below (and including)
      that field are zeroed, effectively allowing a new request.
   */

   /* for maintaining the list of request structures */
   struct  GenericListEntryStruct  RequestListEntry;

#ifdef IP_UCX
   unsigned short  /* channel to client's TCP/IP device */
                   ClientChannel;
#endif

#ifdef IP_NETLIB
   unsigned long  /* socket for client's TCP/IP device */
                  ClientNetLibSocket;
#endif

   unsigned long  /* client's binary IP address */
                  ClientInternetAddress32bit;
   unsigned char  /* internet host/domain name of client */
                  ClientHostName [128],
                  /* numeric host/domain address of client */
                  ClientInternetAddress [32];
   int  /* count of how many times the TCP/IP connection is reused */
        KeepAliveCount;

   /* the rest of this structure can be zeroed with persistent connections */
   unsigned long  RetainAboveZeroBelow;

   boolean  /* "temporary" file, delete it on close */
            DeleteOnClose,
            /* true if a mapping rule identifies this as a "CGIplus" script */
            IsCgiPlusScript,
            /* this request has a "Connection: Keep-Alive" token */
            KeepAliveRequest,
            /* the server has responded with "keep-alive" */
            KeepAliveResponse,
            /* this persistent connection has timed out */
            KeepAliveTimeout,
            /* escapes HTML-forbidden characters as sent to client */
            NetBufferOutputEscapeHtml,
            /* response include an "Expires:" the same as "Last-Modified:" */
            ResponsePreExpired,
            /* "Pragma: no-cache" HTTP request header line encountered */
            PragmaNoCache;

   int  /* */
        AccountingDone,
        /* count of bytes received so far in a POSTed or PUTed body */
        ContentCount,
        /* "Content-Length:" value */
        ContentLength,
        /* */
        ContentTypeLength,
        /* size of buffer space for sys$output from DCL subprocess */
        *DclSysOutputSize,
        /* processing request header, count to find first blank line */
        HeaderConsecutiveNewLineCount,
        /* */
        HttpAcceptLength,
        /* length of error message */
        ErrorMessageLength,
        /* value representation of request HTTP method */
        HttpMethod,
        /* length of request header field */
        HttpAuthorizationLength,
        /* length of request header field */
        HttpRefererLength,
        /* length of request header field */
        HttpUserAgentLength,
        /* size of network read buffer */
        NetReadBufferSize,
        /* size of output buffer */
        OutputBufferSize,
        /* length of path from HTTP header */
        PathInfoLength,
        /* length of query string from HTTp header */
        QueryStringLength,
        /* lenth of resource (path plus query string) from HTTp header */
        ResourceLength,
        /* number of characters up until the blank line */
        RequestHeaderLength,
        /* count of home page file names tried */
        RequestHomePageIndex,
        /* used during HEAD method to determine when header ends */
        ResponseHeaderNewlineCount,
        /* buffer for data length */
        ResponseHeaderDataLength,
        /* length of reponse header */
        ResponseHeaderLength,
        /* numeric equivalent of the "200", "302", etc., HTTP status code */
        ResponseStatusCode,
        /* avoid redirection loops by keeping track of the number */
        RedirectionCount,
        /* VMS status of most recent task */
        TaskStatus;

   unsigned long  /* count of total bytes received from the client */
                  BytesRx,
                  /* count of total bytes sent to the client */
                  BytesTx,
                  /* number of bytes still available in output buffer */
                  OutputBufferRemaining;

   unsigned long  /* 64 bit VMS internal time, start of request processing */
                  BinaryTime [2],
                  /* 64 bit VMS time from "If-Modified-Since:" header */
                  IfModifiedSinceBinaryTime [2],
                  /* length paramater from "If-Modified-Since:" header */
                  IfModifiedSinceLength;

   unsigned short  /* component equivalent of 'BinaryTime' */
                   NumericTime [7];

   unsigned char  /* points to dynamic buffer for body POSTed, PUT, etc. */
                  *ContentBufferPtr,
                  /* used as pointer into above buffer */
                  *ContentPtr,
                  /* buffer space for sys$output from DCL subprocess */
                  *DclSysOutputPtr,
                  /* pointer to heap storage of file URL-style path */
                  *FilePathPtr,
                  /* pointer to error buffer sent to client */
                  *ErrorBufferPtr,
                  /* pointer to error message */
                  *ErrorMessagePtr,
                  /* pointer to explanation when generating error report */
                  *ErrorTextPtr,
                  /* pointer to hidden explanation (often file name) */
                  *ErrorHiddenTextPtr,
                  /* "Accept:" header lines */
                  *HttpAcceptPtr,
                  /* pointer to heap storage of "Authorization:" field */
                  *HttpAuthorizationPtr,
                  /* stream's MIME content type */
                  *HttpContentTypePtr,
                  /* "Cookie:" request header encountered */
                  *HttpCookiePtr,
                  /* pointer to heap storage of "If-Modified-Since:" field */
                  *HttpIfModifiedSincePtr,
                  /* "Pragma:" request header encountered */
                  *HttpPragmaPtr,
                  /* pointer to heap storage of "Referer:" field */
                  *HttpRefererPtr,
                  /* pointer to heap storage of "User-Agent:" field */
                  *HttpUserAgentPtr,
                  /* pointer to heap storage of redirection location path */
                  *LocationPtr,
                  /* pointer to heap storage client-read network buffering */
                  *NetReadBufferPtr,
                  /* pointer to heap storage for output buffering */
                  *OutputBufferPtr,
                  /* pointer used to point into 'Buffer' storage */
                  *OutputBufferCurrentPtr,
                  /* pointer to heap storage for output double-buffering */
                  *OutputDoubleBufferPtr,
                  /* pointer to heap storage of request path */
                  *PathInfoPtr,
                  /* pointer to heap storage of request query string */
                  *QueryStringPtr,
                  /* pointer to requested resource (path and query string) */
                  *ResourcePtr,
                  /* pointer to request header */
                  *RequestHeaderPtr,
                  /* pointer to response header */
                  *ResponseHeaderPtr,
                  /* buffer for data pointer */
                  *ResponseHeaderDataPtr;

   unsigned char  /* GMT-adjusted string equivalent of 'BinaryTime' */
                  GmDateTime [48],
                  /* string reqpresenting the request's HTTP method */
                  HttpMethodName [METHOD_NAME_SIZE],
                  /* authenticated user name from "Authorization:" */
                  RemoteUser [MaxAuthUserNameLength+1],
                  /* storage when authenticating remote user */
                  RemoteUserPassword [MaxAuthPasswordLength+1],
                  /* parsed path storage */
                  RequestFileName [256],
                  /* mapped script name */
                  ScriptName [SCRIPT_NAME_SIZE];

   /* AST function */
   void (*ResponseHeaderAstAddress)(struct RequestStruct*);
   void (*ErrorMessageAstAddress)(struct RequestStruct*);

   /* storage for maintaining the thread's heap storage */
   struct HeapStruct  *HeapHeadPtr;
   struct HeapStruct  *HeapTailPtr;
   int  HeapSize;

   /* */
   struct ContentTypeStruct  ContentType;

   /* I/O status blocks */
   struct AnIOsb  NetReadIOsb,
                  NetWriteIOsb;

   /* RMS structures used for various file processing purposes */
   struct FAB  RequestParseFab;
   struct NAM  RequestParseNam;

   /*****************************************/
   /* pointers to other "object" structures */
   /*****************************************/

   struct DclTaskStruct  *DclTaskPtr;
   struct DescrTaskStruct  *DescrTaskPtr,
                         *DescrTaskPoolPtr;
   struct DirTaskStruct  *DirTaskPtr,
                         *DirTaskPoolPtr;
   struct FileTaskStruct  *FileTaskPtr,
                          *FileTaskPoolPtr;
   struct GraphTaskStruct  *GraphTaskPtr;
   struct HTAdminTaskStruct  *HTAdminTaskPtr;
   struct IsMapTaskStruct  *IsMapTaskPtr;
   struct MenuTaskStruct  *MenuTaskPtr,
                         *MenuTaskPoolPtr;
   struct PutTaskStruct  *PutTaskPtr;
   struct SsiTaskStruct  *SsiTaskPtr,
                         *SsiTaskPoolPtr;
   struct UpdTaskStruct  *UpdTaskPtr;

   /****************************************************/
   /* storage specific to authentication/authorization */
   /****************************************************/

   unsigned long  /* value representation of scheme being used for challenge */
                  AuthChallengeScheme,
                  /* the capability flags of the authorization path */
                  AuthGroupCan,
                  /* the capability flags of the request (ACT ON THESE!) */
                  AuthRequestCan,
                  /* value representation of scheme being used for request */
                  AuthScheme,
                  /* the capability flags of the remote user */
                  AuthUserCan,
                  /* what everybody else can do on that path */
                  AuthWorldCan;

   unsigned char  /* BASIC challenge response header field */
                  *AuthBasicChallengePtr,
                  /* DIGEST challenge response header field */
                  *AuthDigestChallengePtr,
                  /* when digest authenticating, the client-supplied nonce */
                  *AuthDigestNoncePtr,
                  /* when digest authenticating, the client-supplied digest */
                  *AuthDigestResponsePtr,
                  /* when digest authenticating, the client-supplied URI */
                  *AuthDigestUriPtr,
                  /* pointer to authorization group name */
                  *AuthGroupPtr,
                  /* pointer to authorization realm name */
                  *AuthRealmPtr,
                  /* pointer to string of IP-address/username masks */
                  *AuthRestrictListPtr,
                  /* pointer to string of IP-address/username masks */
                  *AuthWorldRestrictListPtr;

   char  /* */
         AuthType [16];
};

#ifdef __ALPHA
#   pragma member_alignment __restore
#endif

/*****************************************************************************/
